/* Mama.c - MandelSpawn master widget */

/*  This file is part of MandelSpawn, a parallel Mandelbrot program for
    the X window system.

    Copyright (C) 1990 Andreas Gustafsson

    MandelSpawn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 1,
    as published by the Free Software Foundation.

    MandelSpawn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License,
    version 1, along with this program; if not, write to the Free 
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* The Mama widget is never realized.  Its purpose is to queue the */
/* requests for calculation coming from the various Ms widgets, */
/* distribute them among the computation servers and return the  */
/* results to the widget that requested them.  It also manages other */
/* common resources such as the colormap. */

/* On some systems, <fcntl.h> may need to be included also */
#include <stdio.h>
#include <errno.h>
extern int errno; /* at least Sony's errno.h misses this */
#include <math.h>

#include <X11/IntrinsicP.h>
#include <X11/Xos.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include "backward.h" /* X11R2 backward compatibility stuff */

#include "MamaP.h"
#include "Ms.h" /* public header of the child widget */

/* names of files containing server hostnames */
#define PERSONAL_SLAVEFILE ".mslaves" 
#ifndef PUBLIC_SLAVEFILE
#define PUBLIC_SLAVEFILE "/usr/local/etc/mslaves"
#endif

#define INITIAL_CHUNKS 1024

#define MAX_WORKPACKET_SIZE 64
#define DATAGRAM_BYTES 1200

/* default timeout in milliseconds per iteration, and constant factor */
#define TIMEOUT_PER_ITER 	40	/* this makes 10 s for 250 iters */
#define TIMEOUT_CONST		3000	/* add 3 seconds for network delays */

/* On the Mach/i386 machines I have tested, floor(0.26) returns -0.5.  Not */
/* good.  Here is a nonportable workaround that's good enough for our */
/* purposes. */
#ifdef BROKEN_FLOOR
#define floor(x) ((double)(long)(x))
#endif

char *getenv();
struct hostent *gethostbyname();

extern XtAppContext thisApp;
extern Display *myDisplay;
extern Screen *myScreen;

static void Initialize(), ClassInitialize(), Realize(), Destroy(), ReExpose(),
  Resize(), GotResult(), Whip(), SetTimeout();

static Boolean SetValues();

/* Defaults */

/* the width/height defaults are never really used */
static Dimension default_width = 400;
static Dimension default_height = 250;
static int default_colors = 250;
static int default_hues = 250;
static char default_spectrum[] = 
 "blue-aquamarine-cyan-medium sea green-forest green-lime green-\
yellow green-yellow-coral-pink-black";
static Bool default_bw = False;
static Bool default_wrap = False;
static Bool default_debug = False;


static XtResource resources[] = 
{
  /* Core resources */
  { XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
      XtOffset(Widget, core.width), XtRDimension, 
      (caddr_t) &default_width },
  { XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
      XtOffset(Widget, core.height), XtRDimension, 
      (caddr_t) &default_height },
  /* Noncore resources */
  { XtNSpectrum, XtCString, XtRString, sizeof(String),
      XtOffset(MamaWidget, mama.spectrum), XtRString,
      default_spectrum },
  { XtNColours, XtCValue, XtRInt, sizeof(int),
      XtOffset(MamaWidget, mama.n_colours), XtRInt,
      (caddr_t) &default_colors },
  { XtNHues, XtCValue, XtRInt, sizeof(int),
      XtOffset(MamaWidget, mama.n_hues), XtRInt,
      (caddr_t) &default_hues },
  { XtNBw, XtCValue, XtRBool, sizeof(Bool),
      XtOffset(MamaWidget, mama.bw), XtRBool,
      (caddr_t) &default_bw },
  { XtNWrap, XtCValue, XtRBool, sizeof(Bool),
      XtOffset(MamaWidget, mama.wrap), XtRBool,
      (caddr_t) &default_wrap },
  { XtNDebug, XtCValue, XtRBool, sizeof(Bool),
      XtOffset(MamaWidget, mama.debug), XtRBool,
      (caddr_t) &default_debug }
};


MamaClassRec mamaClassRec = 
{   /* core fields */
    { 
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Mama",
    /* widget_size		*/	sizeof(MamaRec),
    /* class_initialize		*/      NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	NULL,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* resource_count		*/	XtNumber(resources),
    /* xrm_class		*/	NULL,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	NULL,
    /* expose			*/	NULL,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	NULL
    },
    /* noncore fields */
    {
    /* dummy			*/	0
    }
};

WidgetClass mamaWidgetClass = (WidgetClass) &mamaClassRec;


/* Make a copy of a string in XtMalloc'ed memory (strdup() lookalike) */

static char *
my_strdup(str) 
     char *str;
{ unsigned int length=strlen(str);
  register char *r = (char *) XtMalloc(length+1);
  if(length>0)
    bcopy (str, r, (int) length);
  r[length] = '\0';
  return(r);
}


/* Get the address of a host given either the host name or the address */
/* in nnn.nnn.nnn.nnn notation */

static struct in_addr *
gethostaddrbywhatever(p)
     char *p;
{ struct hostent *hp;

  /* statically allocated return value buffer */
  /* (this is ugly, but gethostbyname() is no better) */
  static struct in_addr nobyinaddr;

  if ((hp=gethostbyname(p)) != NULL)
  { return((struct in_addr *) (hp->h_addr));
  }
  else
  { /* try it as a number nnn.nnn.nnn.nnn */
    if ((nobyinaddr.s_addr=inet_addr(p)) == -1)
    { return(0);
    }
    return(&nobyinaddr);
  }
}


/* Null-terminate the current field in a whitespace-separated list */
/* and return a pointer to the next field or NULL if there is no */
/* next field.  Text after a hash sign is taken as comment and ignored */

static char *
NextField(p) char *p;
{ while(1)
  { register char c = *p;
    switch(c)
    {
    case ' ':
    case '\t':
      *p = '\0';
      while(1)
      { int c = *++p;
	if(c == '\n' || c == '\0') return(0);
	if(c != ' ' && c != '\t') break;
      }
      return(p);
    case '\0':
      return(0);
    case '\n':
    case '#':
      *p = '\0';
      return(0);
    default:
      break;
    }
    p++;
  }
}


/* Read the .mslaves file and set up the slave-related data structures */
void
InitSlaves(w) 
     MamaWidget w;
{ FILE *f;
  int i;
  char *filename, *home;
  unsigned int size=16; /* initial size of slave table */
  char buf[256];
  w->mama.slaves=(slave **) XtMalloc(size*sizeof(slave *));
  home=getenv("HOME");
  if(!home) XtAppError(thisApp, "HOME not set");
  filename=XtMalloc((unsigned)(strlen(home)+1+strlen(PERSONAL_SLAVEFILE)+1));
  strcpy(filename, home); 
  strcat(filename, "/");
  strcat(filename, PERSONAL_SLAVEFILE);
  f=fopen(filename, "r");
  if(!f)
  { f=fopen(PUBLIC_SLAVEFILE, "r");
  }
  if(!f)
  { XtAppError(thisApp, "Could not find .mslaves file");
  }
  i=0;
  while(1)
  { char *p; /* points to current field in .mslaves line */
    char *q; /* points to next field in .mslaves line */
    slave *s;
    struct in_addr *ina;
    unsigned port;

    if(!fgets(buf, sizeof(buf), f)) break;
    if(buf[0]=='\n' || buf[0]=='#') continue;
    p=buf;
    q=NextField(p);
    ina=gethostaddrbywhatever(p);

    if(ina==0)
    { /* I'd like a printf-like XtAppWarning for christmas... */
      static char warn[] = "unknown host in .mslaves, ignored: ";
      char *msg = XtMalloc(sizeof(warn)+strlen(p));
      strcpy(msg, warn);
      strcat(msg, p);
      XtAppWarning(thisApp, msg);
      XtFree(msg);
      continue;
    }

    p=q;
    port=DEFAULT_PORT;
    if(p) /* there is a "port" field */
    { q=NextField(p);
      port=(unsigned) atoi(p);
      if(port==0) /* probably not an integer, and port 0 is bad anyway */
      {	/* I'd like a printf-like XtAppWarning for christmas... */
	static char warn[] = "bad port field in .mslaves, machine ignored: ";
	char *msg = XtMalloc(sizeof(warn)+strlen(p));
	strcpy(msg, warn);
	strcat(msg, p);
	XtAppWarning(thisApp, msg);
	XtFree(msg);
	continue;
      }
    }

    if(q)
    { XtAppWarning(thisApp, "trailing junk in .mslaves");
    }

    if(i>=size) 
    { size *= 2;
      w->mama.slaves=
	(slave **) XtRealloc((char *) w->mama.slaves, size*sizeof(slave *));
    }
    s=(slave *) XtMalloc(sizeof(slave));  
    bcopy((char *) ina, (char *) &s->name.sin_addr, sizeof(struct in_addr));
    s->name.sin_family = AF_INET;
    s->name.sin_port = htons(port);
    
    s->name_string = my_strdup(buf);
    s->mi_count = 0L;
    s->has_timeout = False;
    s->n_timeouts = s->n_packets=s->n_late_packets=0;
    s->no = i;
    s->mama = w;
    s->disabled =0;
    s->timeout = w->mama.n_hues * TIMEOUT_PER_ITER + TIMEOUT_CONST;
    w->mama.slaves[i] = s;
    i++;
  }
  w->mama.n_slaves=i;
}


/* Set up the colour map */

static void 
  MsColourSetup(w)
MamaWidget w;
{ unsigned int i;
  unsigned int ncolours;
  unsigned int nhues;
  unsigned int nstops;
  unsigned long *phpixels;
  unsigned long *pixels;
  XColor *stops;
  XColor *colourset;
  unsigned stops_allocated=4; /* initial allocation of colour stops */
  char *p;
  stops=(XColor *) XtMalloc(stops_allocated * sizeof(XColor));
  /* parse the "spectrum" resource */
  p=w->mama.spectrum;
  i=0;
  while(1)
  { char *end;
    XColor dummy;
    end=index(p, '-');
    if(end) *end = '\0';
    if(i >= stops_allocated)
    { stops_allocated *= 2;
      stops=(XColor *)
	XtRealloc((char *) stops, stops_allocated * sizeof(XColor));
    }
    if(! XLookupColor(myDisplay, DefaultColormapOfScreen(myScreen), p,
		     &stops[i], &dummy))
    { static char *err = "unrecognized colour in spectrum: ";
      char *msg = XtMalloc(sizeof(err)+strlen(p));
      strcpy(msg, err);
      strcat(msg, p);
      XtAppError(thisApp, msg);
      XtFree(msg); /*NOTREACHED*/
    }
    i++;
    if(end) p=end+1; 
    else break;
  }
  nstops=i;
  ncolours=w->mama.n_colours;
  nhues=w->mama.n_hues;

  /* physical pixel mapping (from colour number to pixel value, */
  /* "ncolours" entries) */
  phpixels=
    (unsigned long *) XtMalloc(ncolours*sizeof(unsigned long));
  /* logical pixel mapping (from iteration count to pixel value, */
  /* "nhues" entries) */
  pixels=
    (unsigned long *) XtMalloc(nhues*sizeof(unsigned long));

  w->mama.my_colormap = DefaultColormapOfScreen(myScreen);
  
  if(! XAllocColorCells(myDisplay, w->mama.my_colormap,
		       False, NULL, 0,  
		       phpixels, ncolours))
  { /* allocation our of the default colormap failed, so we */
    /* try to allocate one of our own */
    XColor dummy;
    w->mama.my_colormap =
      XCopyColormapAndFree(myDisplay, w->mama.my_colormap);
    if(! w->mama.my_colormap)
      XtAppError(thisApp, "XCopyColormapAndFree failed");
#ifdef MENU
    /* make sure that this colourmap also has black and white pixels */
    /* for use by the popup menu, if any.  Don't care about errors here. */
    (void) XAllocNamedColor(myDisplay, w->mama.my_colormap,
			    "white", &dummy, &dummy);
    (void) XAllocNamedColor(myDisplay, w->mama.my_colormap,
			    "black", &dummy, &dummy);
#endif
    /* retry the allocation */
    if(! XAllocColorCells(myDisplay, w->mama.my_colormap,
			  False, NULL, 0,  
			  phpixels, ncolours))
    { XtAppError(thisApp,
"too many colours for this screen.  Please specify a smaller number\n\
of colours with the -colours option, or use -bw to run in black and\n\
white.");
    }
  }

  colourset=(XColor *) XtMalloc(ncolours*sizeof(XColor));
  for(i=0; i<ncolours; i++)
  { double flstop, stop, fpart;
    int istop, jstop;
    flstop=(double) i / (double) (ncolours-1) * (double) (nstops-1);
    /* calculate the number of the next lower stop */
    /* should be ==nstops-1 only for i==ncolours-1 */
    stop=floor(flstop);
    fpart=flstop-stop;
    /* calculate the indices of the two stops to interpolate between */
    /* If we are at the last stop, interpolate between it and itself */
    /* to avoid accesses past the end of the array (the value will be */
    /* multiplied by zero anyway) */
    istop=(int) stop;
    jstop=(istop+1 >= nstops ? nstops-1 : istop+1);

    colourset[i].pixel=phpixels[i];
    colourset[i].red=
      floor((1-fpart)*stops[istop].red + fpart*stops[jstop].red);
    colourset[i].green=
      floor((1-fpart)*stops[istop].green + fpart*stops[jstop].green);
    colourset[i].blue=
      floor((1-fpart)*stops[istop].blue + fpart*stops[jstop].blue);
    colourset[i].flags=DoRed|DoGreen|DoBlue;
  }

  if(w->mama.wrap)
  { /* Wrap the spectrum around several times if needed to get the */
    /* desired number of "hues" from a limited number of "colours" */
    for(i=0; i<nhues-1; i++)
    { pixels[i] = phpixels[i % ncolours];
    }
  }
  else
  { /* Don't wrap; give the same colour to several consecutive iteration */
    /* values */
    int divfactor = (nhues + ncolours-1) / ncolours;
    for(i=0; i<nhues-1; i++)
    { pixels[i] = phpixels[i / divfactor];
    }
  }
  /* ...but make sure the last colour is what the user specified. */
  pixels[nhues-1] = phpixels[ncolours-1];
  w->mama.pixels=pixels;
  
  XStoreColors(myDisplay, w->mama.my_colormap, 
	       colourset, ncolours);
  XtFree((char *) colourset);
}


/* Set up for black-and-white operation */

static void 
  MsBwSetup(w)
MamaWidget w;
{ unsigned long *pixels;

  unsigned niter=w->mama.n_colours;
  register int i;
  
  pixels=w->mama.pixels=
    (unsigned long *) XtMalloc(niter*sizeof(unsigned long));

  /* Use alternating black/white bands */
  for(i=0; i<niter; i++)
  { pixels[i] = (niter - i) & 0x01 ?
      BlackPixelOfScreen(myScreen) : WhitePixelOfScreen(myScreen);
  }
}


/* Callback function to be called when data arrives from a slave */

void MsSocketInputCallback(client_data, source, id)
     caddr_t client_data; int *source; XtInputId *id;
{ Message msg;
  int fromlen;
  MamaWidget w=(MamaWidget) client_data;
  if(*id != w->mama.input_id || *source != w->mama.my_socket) 
  { XtAppError(thisApp, "unexpected input from slave");
  }
  /* This previously used read(), but some non-BSD TCP/IP implementations */
  /* don't allow it to be used with connectionless sockets.  Also, */
  /* while the Sun implementation does allow for a null pointer */
  /* for the "from" argument in recvfrom(), the "fromlen" argument */
  /* may not be a null pointer. */
  if(recvfrom(w->mama.my_socket, (char *) &msg, sizeof(msg),
	      0, (struct sockaddr *) 0, &fromlen) == -1)

  {
  /* If you get spurious "operation would block" errors, enable this and */
  /* send a bug report to gson@niksula.hut.fi */
#if 0 
    if(errno != EWOULDBLOCK)
#endif
    { perror("read");
      XtAppError(thisApp, "slave input socket read");
    }
  }
  else
  { GotResult(w, &msg);
  }
}


/* Initialize the widget */
static void
Initialize(request, new)
     MamaWidget request;
     MamaWidget new;
{ /* we haven't made any popups yet */
  new->mama.n_popups_created=0;

  /* set up the chunk index */
  new->mama.max_chunks=INITIAL_CHUNKS;
  new->mama.n_chunks=0;
  new->mama.chunks=(chunk **) XtMalloc(new->mama.max_chunks * sizeof(chunk *));
  
  /* set up the chunk queues */
  new->mama.to_draw.prev=new->mama.to_draw.next= &new->mama.to_draw;
  new->mama.drawn.prev=new->mama.drawn.next= &new->mama.drawn;

  /* read the .mslaves file and set new->mama.n_slaves and .slaves */
  InitSlaves(new);

  /* make a socket for communicating with the slaves */
  if((new->mama.my_socket=
      socket(AF_INET, SOCK_DGRAM, 0)) < 0) 
    XtAppError(thisApp, "socket");
  
  /* make sure we don't block reading the socket */  
  if(fcntl(new->mama.my_socket, F_SETFL, FNDELAY) == -1)
    XtAppError(thisApp, "unblocking socket");

  /* pass a pointer to the master widget as "client data" */
  new->mama.input_id=
    XtAppAddInput(thisApp,
		  new->mama.my_socket,
		  (caddr_t) XtInputReadMask,
		  MsSocketInputCallback,
		  (caddr_t) new);
  
  new->mama.pid=getpid();
  new->mama.sequence=0;
  
  if(new->mama.n_colours > new->mama.n_hues)
  { /* we have more colours than iterations, throw some away */
    new->mama.n_colours = new->mama.n_hues;
  }
  if(new->mama.bw)
    goto force_bw;

  /* set new->mama.pixels and the colour map, if applicable */
  switch(DefaultVisualOfScreen(myScreen)->class)
  {
  case PseudoColor:
    MsColourSetup(new); break;
  case StaticGray:
  force_bw:
    MsBwSetup(new); break;
  default:
    XtAppError(thisApp, "unsupported visual type");
  }
}


/* Delete the chunk "c" from anywhere in the queue */
/* Don't deallocate the chunk yet because duplicated messages may still */
/* refer to it */

void
DeleteQueue(c)
     chunk *c;
{ c->next->prev=c->prev;
  c->prev->next=c->next;
  c->next=c->prev=NULL; /* just for easier debugging */
}


/* Add a chunk to the head of the queue */

void
AddQueue(q, c) chunk *q; chunk *c;
{ c->prev=q->prev;
  c->next=q;
  q->prev->next=c;
  q->prev=c;
}


/* Test the emptiness of a chunk queue */

Bool
EmptyQueue(q)
     chunk *q;
{ return(q->next==q && q->prev==q);
}


/* Return the head of a chunk queue */

chunk *
HeadOfQueue(q)
     chunk *q;
{ return(q->next);
}


/* Callback to be called whenever a slave times out */

void 
TimeoutCallback(client_data, id) 
     caddr_t client_data;
     XtIntervalId *id;
{ slave *s=(slave *) client_data;
  s->has_timeout=False;
  s->n_timeouts++;
  Whip(s->mama, s);
}


/* Remove a timeout */

void 
UnsetTimeout(s)
     slave *s;
{ if(s->has_timeout)
    XtRemoveTimeOut(s->timer);
  s->has_timeout=False;
}


/* Set a timeout for a slave */
/* This is done when the slave is whipped */

static void
SetTimeout(s)
     slave *s;
{ /* first make sure there isn't a timeout already */
  UnsetTimeout(s);
  s->timer=XtAppAddTimeOut(thisApp, s->timeout,
			   TimeoutCallback, (caddr_t) s);
  s->has_timeout=True;
}


/* Stop the slaves */

void
StopSlaves(w)
     MamaWidget w;
{ int i;
  for(i=0; i<w->mama.n_chunks; i++)
  { chunk *c = w->mama.chunks[i];
    XtFree((char *) (c->client_data));
    XtFree((char *) (c->slave_data));
    XtFree((char *) c);
  }
  /* don't shrink the chunk index array; we probably need it again */
  w->mama.n_chunks=0;
  w->mama.sequence++;
  w->mama.drawn.prev=w->mama.drawn.next= &w->mama.drawn;
}

/* Put the specified slave to work, or stop all slaves if all work is done */

static void 
Whip(w, s)
     MamaWidget w;
     slave *s;
{ struct
  { WhipMessage m;
    char data[MAX_WORKPACKET_SIZE];
  } mm; /* buffer for building the message to send */
  chunk *c;
  if(s->disabled)
    return;
  if(EmptyQueue(&w->mama.to_draw))
  { StopSlaves(w);
  }
  else /* queue not empty */
  { MsWidget msw;
    c=HeadOfQueue(&w->mama.to_draw);
    msw=c->wid;
    mm.m.header.magic=htons(MAGIC);
    mm.m.header.type=htons(WHIP_MESSAGE);
    mm.m.header.version=htons(VERSION + (int) w->mama.debug);
    mm.m.header.format=htons(DATA_FORMAT);
    mm.m.id.pid=w->mama.pid;
    mm.m.id.seq=w->mama.sequence;
    mm.m.id.slave_no=s->no;
    mm.m.id.chunk_no=c->no;

    if(c->slave_datalen > MAX_WORKPACKET_SIZE)
      XtAppError(thisApp, "work packet too large");
    bcopy(c->slave_data, mm.m.data, c->slave_datalen);

    if(sendto(w->mama.my_socket, 
	      (char *)&mm, sizeof(mm.m)+c->slave_datalen,
	      0, (struct sockaddr *) &s->name,
	      (int) sizeof(s->name)) == -1)
    { XtAppWarning(thisApp, "error sending datagram, machine disabled");
      s->disabled=1; /* consider this slave unusable */
    }

    SetTimeout(s); 
    /* move the chunk from the head to the tail of the queue */
    DeleteQueue(c);
    AddQueue(&w->mama.to_draw, c);
  }
}


/* Handle a reply from a slave */

static void 
GotResult(mama, msg) 
     MamaWidget mama;
     Message *msg;
{ MsWidget w;
  slave *s;
  chunk *c;
  
  unsigned int pid=msg->reply.id.pid;
  unsigned int seqno=msg->reply.id.seq;
  unsigned int chunkno=msg->reply.id.chunk_no;
  unsigned int slaveno=msg->reply.id.slave_no;
  int late;
  
  if(pid != mama->mama.pid) return;
  if(slaveno >= mama->mama.n_slaves) return;
  s= mama->mama.slaves[slaveno];
  s->n_packets++;
  if(seqno != mama->mama.sequence) 
  { s->n_late_packets++;
    return;
  }
  if(chunkno >= mama->mama.n_chunks) return;
  c= mama->mama.chunks[chunkno];
  
  UnsetTimeout(s);

  w=c->wid;

  /* the chunk is too late if the window has gone away or it has */
  /* been drawn already */
  late = (!w || c->drawn);

  if(late) /* ignore the message if the chunk arrived too late */
  { s->n_late_packets++;
  }
  else
  { DeleteQueue(c);
    AddQueue(&mama->mama.drawn, c);
    c->drawn=True;
  }
  /* Put the slave to work again, or stop all slaves if there is */
  /* no more work */
  Whip(mama, s); 
  if(!late)
  { Draw(w, c->client_data, 
	 (caddr_t) &(msg->reply.data), mama->mama.pixels);
    s->mi_count += ntohl(msg->reply.mi_count);
  }
}


/* This function is called by the Ms widget for each chunk it */
/* wants to be calculated */

void 
DispatchChunk(m, w, client_data, client_datalen, slave_data, slave_datalen)
     MamaWidget m;
     MsWidget w;
     caddr_t client_data;
     unsigned int client_datalen;
     caddr_t slave_data;
     unsigned int slave_datalen;
{ int chunkno=m->mama.n_chunks++;
  chunk *c=(chunk *) XtMalloc(sizeof(chunk));
  c->client_data=(caddr_t) XtMalloc(client_datalen);
  bcopy(client_data, c->client_data, client_datalen);
  c->slave_data=(caddr_t) XtMalloc(slave_datalen);
  bcopy(slave_data, c->slave_data, slave_datalen);
  c->slave_datalen=slave_datalen;
  c->wid=w;
  c->drawn=False;
  c->no=chunkno;
  /* grow the chunk index if necessary */
  if(chunkno >= m->mama.max_chunks)
  { m->mama.max_chunks *= 2;
    m->mama.chunks=(chunk **) XtRealloc((char *) m->mama.chunks,
				       m->mama.max_chunks*sizeof(chunk *));
  }
  m->mama.chunks[chunkno]=c;
  AddQueue(&m->mama.to_draw, c);
}


/* Put all the slaves to work */

void WhipAllSlaves(w)
 MamaWidget w;
{ int i;
  for(i=0; i<w->mama.n_slaves; i++)
  { Whip(w, w->mama.slaves[i]);
  }
}


/* Print some performance statistics about the slaves */

void SlaveStatistics(w)
     MamaWidget w;
{ int i;
  int active=0;
  unsigned long mi_tot=0;
  printf("\n%-22s %10s %10s %10s %10s\n",
	 "Host", "iterations", "packets", "timeouts", "late");
  for(i=0; i<w->mama.n_slaves; i++)
  { slave *s= w->mama.slaves[i];
    printf("%-22s %10lu %10u %10u %10u\n",
	   s->name_string, 
	   s->mi_count, s->n_packets, s->n_timeouts, s->n_late_packets);
    if(s->mi_count) active++;
    mi_tot += s->mi_count;
  }
  printf("%d servers, %d active\n", w->mama.n_slaves, active);
  fflush(stdout);
}


/* Update widget resources (no changeable resources so far) */

static Boolean SetValues(current, request, new)
     MamaWidget current, request, new;
{ return(False);
}


/* Die */

void Shutdown(w)
     MamaWidget w;
{ XCloseDisplay(XtDisplay(w));
  exit(0);
}


/* This callback is called whenever a popup child dies */

void
PostMortem(w, client_data, call_data)
     MsWidget w;
     caddr_t client_data;
     caddr_t call_data;
{ MamaWidget ma=(MamaWidget) client_data;
  chunk *c;
  chunk *next_c;
  /* Remove all chunks of the dead widget from the work queue */
  /* and nullify their widget pointer fields so that late packets */
  /* won't reference the nonexistent widget */
  for(c=ma->mama.to_draw.next; c != &ma->mama.to_draw; c=next_c)
  { /* need to use a temporary variable because */
    /* relinking the chunk fouls up the .next field */
    next_c=c->next; 
    if(c->wid == w)
    { c->wid = NULL;
      DeleteQueue(c);
      AddQueue(&ma->mama.drawn, c);
    }
  }
  /* If our last child is dying we have nothing to live for; */
  /* commit suicide */
  if(ma->core.num_popups == 1)
    Shutdown(ma);
}


/* Pop up a new Ms window, making it a child of the Mama. */
/* The caller must make sure there is room in the "shell_args" array */
/* for one more arg. */

void PopupAnother (w, shell_args, num_shell_args, args, num_args)
     MamaWidget w;
     ArgList shell_args;
     Cardinal num_shell_args;
     ArgList args;
     Cardinal num_args;
{ Widget the_widget;
  Widget the_shell;
  char popup_name[32]; /* unique name for each popup widget */

  sprintf(popup_name, "ms_%d", ++w->mama.n_popups_created);

  /* make the popup use our colormap */
  XtSetArg(shell_args[num_shell_args], XtNcolormap, w->mama.my_colormap);
  num_shell_args++;

  the_shell=
    XtCreatePopupShell(popup_name, topLevelShellWidgetClass, w, 
		       shell_args, num_shell_args);
  
  the_widget=XtCreateManagedWidget("view",
				   (WidgetClass) msWidgetClass,
				   (Widget) the_shell, 
				   args, num_args);
  XtAddCallback(the_widget, XtNdestroyCallback, PostMortem, (caddr_t) w);
		
  XtPopup(the_shell, XtGrabNone);
}


/* Destroy the widget */

static void
Destroy (w)
  MamaWidget w;
{
}


/* The Ms widget needs to know the maximums message size, */
/* how may colours there are, etc.  Make that information */
/* available through public functions */

unsigned
MaxMessageSize()
{ return(DATAGRAM_BYTES-sizeof(ReplyHeader));
}

unsigned 
MaxIterations(w)
     MamaWidget w;
{ return(w->mama.n_hues);
}

unsigned MamaHeight(w)
     MamaWidget w;
{ return(w->core.height);
}

unsigned MamaWidth(w)
     MamaWidget w;
{ return(w->core.width);
}

unsigned 
MamaColormap(w)
     MamaWidget w;
{ return(w->mama.my_colormap);
}
